"""
Daline体系中所有的单元类型定义,unit负责定义接口,frame负责框架和执行。所有业务单元均派生于以下基础单元之一
UnitBase: 单元基类,作为PU,TU等单元父类,不直接作为业务单元父类
ProcessUnit: 常驻业务服务的基类,派生常驻型后台服务
TaskUnit: 一次性或定期任务型脚本基类,做一次性或定期数据处理任务,不直接对外提供服务接口
TradeStrategyUnit: 量化交易策略单元基类,包含策略回测,实盘托管等接口定义
任何业务单元仅支持从预设单元类型中派生,单实例。按规范要求实现业务逻辑并在规定接口中返回数据结果,并要按规范要求编写docstring,代码编写及各类命名遵循编码规范,pylint无报错。
"""
import os
import threading
import time
import redis
import requests
import random
import queue
import json
import datetime
# import logging
from functools import wraps
from logging import Logger
from abc import ABC, abstractmethod
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.orm import Session
from sqlalchemy.orm.decl_api import registry


class Base(DeclarativeBase):
    pass


class DalEncoder(json.JSONEncoder):
    def default(self, obj):
        # if isinstance(obj.__class__, DeclarativeMeta):
        if isinstance(obj, DeclarativeBase):
            # an SQLAlchemy class
            fields = {}
            for field in [x for x in dir(obj) if not x.startswith('_') and x != 'metadata']:
                data = obj.__getattribute__(field)
                try:
                    if isinstance(data, registry):
                        continue
                    elif isinstance(data, datetime.datetime) or \
                            isinstance(data, datetime.date) or \
                            isinstance(data, datetime.time):
                        fields[field] = data.isoformat()
                    else:
                        # this will fail on non-encodable values, like other classes
                        json.dumps(data, cls=DalEncoder)
                    fields[field] = data
                except TypeError:
                    fields[field] = None
            # a json-encodable dict
            return fields
        # elif isinstance(obj, DeclarativeBase):
        elif isinstance(obj, datetime.datetime) or \
                isinstance(obj, datetime.date) or \
                isinstance(obj, datetime.time):
            return obj.isoformat()
        return json.JSONEncoder.default(self, obj)


def alchemy_to_pyobj(alobj):
    """
    alchemy对象转python对象
    """
    if isinstance(alobj, list):
        ret = []
        if len(alobj) != 0:
            attributes = [d for d in dir(alobj[0])
                          if not d.startswith('_') and d not in ['metadata', 'registry']]
            for obj in alobj:
                ret_obj = {}
                for item in attributes:
                    ret_obj[item] = getattr(obj, item)
                    if isinstance(ret_obj[item], list) or isinstance(ret_obj[item], DeclarativeBase):
                        ret_obj[item] = alchemy_to_pyobj(ret_obj[item])
                ret.append(ret_obj)
    else:
        ret = {}
        attributes = [d for d in dir(alobj)
                      if not d.startswith('_') and d not in ['metadata', 'registry']]
        for item in attributes:
            ret[item] = getattr(alobj, item)
            if isinstance(ret[item], list) or isinstance(ret[item], DeclarativeBase):
                ret[item] = alchemy_to_pyobj(ret[item])
    return ret
    # js = json.dumps(alobj, cls=DalEncoder)
    # return json.loads(js)


def db_session(obj_func):
    """
    自动获取并释放数据库session给需要使用的函数
    """
    @wraps(obj_func)
    def wrap_dbuse_fun(unit_ins, *args, **kwargs):
        with Session(unit_ins.sql_engine) as session:
            # unit_ins.logger.info(f"[AAA]with in function:{obj_func.__name__}")
            try:
                kwargs['session'] = session
                ret = obj_func(unit_ins, *args, **kwargs)
            except Exception as e:
                # unit_ins.logger.info(
                #     f"[AAA]exept out function:{obj_func.__name__}")
                raise e
            # unit_ins.logger.info(
            #     f"[AAA]return out function:{obj_func.__name__}")
            return ret
    return wrap_dbuse_fun


class UnitBase(ABC):
    """
    所有单元基类,主要处理日志、接口路由等基础功能

    Attributes:
        _logger (logger): 日志实例
    """

    def __init__(self):
        """
        初始化
        """
        self._curcall_request = None
        self._logger = None
        self._logger_fresh_fun = None
        self._sql_config = {}
        self._sql_engine = None
        self._redis_config = {}
        self._redis_pool = None
        self._folder_temp = ''
        self._folder_nfs = ''
        self._folder_http = ''
        self._http_base = ''
        self._httpstore_base = ''
        self._request_func = None
        self._request_glb_func = None
        self._reg_request = {}
        self.GLB_REQUEST_START = 5000
        self._reg_glb_request = {}
        # self._subscription_func = None
        # self._unsubscription_func = None
        self._reg_subscription = {}
        self._token = {}
        self._ws_udefine_emit_fun = None

    @abstractmethod
    def test(self) -> bool:
        """
        各业务单元自定义单元测试部分

        Returns:
            True: 自动测试成功
            False: 自动测试失败
        """
        return True

    @classmethod
    def i18n_retinfo(cls, ret_info) -> str:
        """
        国际化返回提示信息,负责单元实例范围内的返回信息国际化,远期将各语言版本分别用对应文件加载
        """
        i18n = os.getenv('I18N', 'en')
        i18n_all = {
            'cn': {
                'success': "成功"
            }
        }
        return i18n_all.get(i18n, {}).get(ret_info, ret_info)

    @property
    def curcall_request(self) -> any:
        """
        获取当前调入的request实例
        """
        return self._curcall_request
        # return logging.getLogger()

    @curcall_request.setter
    def curcall_request(self, value):
        """
        设置当前调入的request实例
        """
        self._curcall_request = value

    @property
    def logger(self) -> Logger:
        """
        使用logger实例
        """
        return self._logger
        # return logging.getLogger()

    @logger.setter
    def logger(self, value):
        """
        设置本单元所用logger实例

        Parameters:
            logger (logger): logger实例
        """
        self._logger = value

    def set_logger_fresh_fun(self, fun):
        """
        设置模块日志刷新设置函数,引入会打印日志的第三方库时调用logger_fresh_set()接管

        Parameters:
            fun (fun): 日志刷新设置函数
        """
        self._logger_fresh_fun = fun

    def logger_fresh_set(self):
        """
        引入会打印日志的第三方库时调用,做日志打印接管

        Parameters:
            fun (fun): 日志刷新设置函数
        """
        self._logger_fresh_fun()

    @property
    def sql_config(self):
        """
        数据库连接配置
        """
        return self._sql_config

    def set_sql_config(self, config):
        """
        设置数据库连接引擎

        Parameters:
            engine (sql engine): 数据库连接引擎
        """
        self._sql_config = config

    @property
    def sql_connection(self):
        """
        使用sql connection实例
        """
        if self._sql_engine:
            return self._sql_engine.connect()
        else:
            return None

    @property
    def sql_engine(self):
        """
        数据库连接引擎实例
        """
        return self._sql_engine

    def set_sql_engine(self, engine):
        """
        设置数据库连接引擎

        Parameters:
            engine (sql engine): 数据库连接引擎
        """
        self._sql_engine = engine

    @property
    def redis_config(self):
        """
        redis连接配置
        """
        return self._redis_config

    def set_redis_config(self, config):
        """
        设置redis连接引擎配置

        Parameters:
            engine (redis engine): 设置redis连接引擎配置
        """
        self._redis_config = config

    @property
    def redis_conn(self):
        """
        数据库连接引擎实例
        """
        if self._redis_pool:
            return redis.Redis(connection_pool=self._redis_pool)
        else:
            return None

    def set_redis_pool(self, redis_pool):
        """
        设置redis连接池

        Parameters:
            redis_pool (redis pool): redis连接池
        """
        self._redis_pool = redis_pool

    @property
    def folder_temp(self) -> str:
        """
        获取临时存储目录
        """
        return self._folder_temp

    @folder_temp.setter
    def folder_temp(self, path) -> str:
        """
        设置本单元所用临时存储目录,仅做临时缓存使用,仅本单元可见,单元重启后数据可能丢失

        Parameters:
            path (string): 临时存储目录的绝对路径
        """
        self._folder_temp = path

    @property
    def folder_nfs(self) -> str:
        """
        获取单元用NFS存储目录
        """
        return self._folder_nfs

    @folder_nfs.setter
    def folder_nfs(self, path) -> str:
        """
        设置本单元本版本所用NFS存储目录,永久存储,同名业务单元的多个实例共享,单元重启后数据不丢失

        Parameters:
            path (string): 存储目录的绝对路径
        """
        self._folder_nfs = path

    @property
    def folder_nfs_unit(self) -> str:
        """
        获取本单元NFS存储目录,本单元各版本存储目录都在本目录下
        """
        return os.path.split(os.path.split(self.folder_nfs)[0])[0]+"/"

    @property
    def folder_nfs_author(self) -> str:
        """
        获取本作者NFS存储目录,各单元存储目录都在本作者目录下
        """
        return os.path.split(os.path.split(self.folder_nfs_unit)[0])[0]+"/"

    def set_http_base(self, base):
        """
        设置本单元所用web存储目录的基础链接地址

        Parameters:
            base (string): 基础链接地址
        """
        self._http_base = base

    def set_httpstore_base(self, base):
        """
        设置本单元所用web存储目录的httpstore链接地址,给本space下web工程使用

        Parameters:
            base (string): 基础链接地址
        """
        self._httpstore_base = base

    @property
    def folder_http(self) -> str:
        """
        获取本单元本版本用http存储目录
        """
        return self._folder_http

    @property
    def folder_http_unit(self) -> str:
        """
        获取单元用http存储目录
        """
        return os.path.split(os.path.split(self.folder_http)[0])[0]+"/"

    @property
    def folder_http_author(self) -> str:
        """
        获取本作者用http存储目录
        """
        return os.path.split(os.path.split(self.folder_http_unit)[0])[0]+"/"

    @folder_http.setter
    def folder_http(self, path):
        """
        设置本单元所用web存储目录,永久存储,同名业务单元的多个实例共享,单元重启后数据不丢失,有URL链接地址,可通过http链接公开访问到

        Parameters:
            path (string): 存储目录的绝对路径
        """
        self._folder_http = path

    def http_link(self, path) -> str:
        """
        获取http存储中path对应公开访问的http链接

        Parameters:
            path (string): 包含folder_http的绝对路径
        """
        return path.replace(self._folder_http, self._http_base)

    def httpstore_link(self, path) -> str:
        """
        获取http存储中path对应/httpstore目录路径,用于给本空间的web链接,链接在web工程中再经过代理访问到实际地址

        Parameters:
            path (string): 包含folder_http的绝对路径
        """
        return path.replace(self._folder_http, self._httpstore_base)

    def register_request(self, unit, version, fidx):
        """
        注册对其他单元接口的调用
        Parameters:
            unit (string): 其他单元名,如'pu_fakeif'
            version (int): 其他单元版本号,如1.0.1
            fidx (string): 调用接口,如'example'
            method (string): 调用方式,如'get','post'等
        Returns:
            调用句柄
        """
        handle = len(self._reg_request)
        self._reg_request[handle] = {}
        self._reg_request[handle]['unit'] = unit
        self._reg_request[handle]['version'] = version
        self._reg_request[handle]['fidx'] = fidx
        return handle

    @property
    def registerd_request(self) -> dict:
        """
        所有对其他单元调用注册
        """
        return self._reg_request

    def register_glb_request(self, unit, version, fidx):
        """
        注册全局系统基础服务的单元接口的调用,全局服务控制各环境,各space的部署及开发者账户等全局信息
        Parameters:
            unit (string): 其他单元名,如'pu_sys_unitcenter'
            version (int): 其他单元版本号,如1.0.1
            fidx (string): 调用接口,如'example'
            method (string): 调用方式,如'get','post'等
        Returns:
            调用句柄
        """
        handle = len(self._reg_glb_request)
        self._reg_glb_request[handle] = {}
        self._reg_glb_request[handle]['unit'] = unit
        self._reg_glb_request[handle]['version'] = version
        self._reg_glb_request[handle]['fidx'] = fidx
        return handle+self.GLB_REQUEST_START

    @property
    def registerd_glb_request(self) -> dict:
        """
        所有全局单元调用注册
        """
        return self._reg_glb_request

    def set_request_func(self, func):
        """
        框架设置单元调用其他单元的入口函数
        Parameters:
            func (method): 入口函数
        """
        self._request_func = func

    def set_glb_request_func(self, func):
        """
        框架设置单元调用全局基础服务单元的入口函数
        Parameters:
            func (method): 入口函数
        """
        self._request_glb_func = func

    def call_unit(self, handle, index) -> tuple:
        """
        调用其他单元接口
        Parameters:
            func (method): 入口函数
        """
        if handle not in self._reg_request:
            raise Exception(f"Handle[{handle}] not registered!")
        return self._request_func(self._reg_request[handle], index)

    def call_glb_unit(self, handle, index) -> tuple:
        """
        调用全局基础服务单元接口
        Parameters:
            func (method): 入口函数
        """
        handle -= self.GLB_REQUEST_START
        if handle not in self._reg_glb_request:
            raise Exception(f"Handle[{handle}] not registered!")
        return self._request_glb_func(self._reg_glb_request[handle], index)

    @property
    def token(self) -> dict:
        """
        开发者token,仅供特定场景使用,不要直接使用,调用其他单元通过registerd_request和registerd_glb_request
        """
        return self._token

    def set_token(self, token) -> dict:
        """
        设置开发者token
        """
        self._token = token

    def current_env(self) -> str:
        """
        当前所处的环境
        """
        return os.getenv('DAL_ENV', 'dev')

    def in_customer_env(self) -> bool:
        """
        是否在客户生产环境
        """
        return bool(int(os.getenv('IN_CUSTOMER_ENV', '0')))

    def in_k8s(self) -> bool:
        """
        是否运行在集群中
        """
        return bool(int(os.getenv('IN_K8S', '0')))

    def set_ws_udefine_emit_fun(self, fun):
        """
        框架设置单元通过websocket发送单元自定义数据的回调接口
        Parameters:
            fun (method): 入口函数
        """
        self._ws_udefine_emit_fun = fun

    def ws_udefine_emit(self, unit, version, data):
        """
        通过websocket发送单元自定义数据
        """
        if self._ws_udefine_emit_fun:
            self._ws_udefine_emit_fun(unit, version, data)


class ProcessUnit(UnitBase):
    """
    常驻业务单元基类
    """

    def __init__(self):
        """
        初始化
        """
        super().__init__()
        self._task_update_progress_func = None
        self._get_sub_byds_func = None
        self._get_sub_byds_withsid_func = None

    @abstractmethod
    def custom_init(self) -> bool:
        """
        各业务单元自定义初始化部分,初始化完成后单元进入正常运行状态,如果出错会自动尝试重新调用此函数再进行初始化直至成功

        Returns:
            True: 初始化完成
            False: 初始化中有错误,会自动重新调用
        """
        return True

    @classmethod
    def i18n_retinfo(cls, ret_info) -> str:
        """
        国际化返回提示信息,负责单元范围内的返回信息国际化,远期将各语言版本分别用对应文件加载
        """
        sup_ret_info = super().i18n_retinfo(ret_info)
        i18n = os.getenv('I18N', 'en')
        i18n_all = {
            'cn': {
            }
        }
        return i18n_all.get(i18n, {}).get(ret_info, sup_ret_info)

    def ws_connected(self, _sid):
        """
        websocket连入单元
        """
        return

    def ws_disconnected(self, _sid):
        """
        websocket断连
        """
        return

    def ws_subscribed(self, _sid, _if):
        """
        websocket订阅数据
        """
        return

    def ws_unsubscribed(self, _sid, _if):
        """
        websocket取消订阅数据
        """
        return

    def ws_disconnected(self, _sid):
        """
        websocket断连
        """
        return

    def ws_udefine(self, _sid, _caller, _data):
        """
        单元开发者自定义websocket数据推入
        """
        return

    def set_task_update_progress_func(self, func):
        """
        框架设置更新任务进度回调函数
        Parameters:
            func (method): 入口函数
        """
        self._task_update_progress_func = func

    def task_update_progress(self, progress):
        """
        更新任务进度,progress为float型,0~100
        """
        if self._task_update_progress_func:
            self._task_update_progress_func(progress)

    def set_get_sub_byds_func(self, func):
        """
        框架设置获取按ds获取订阅列表函数
        Parameters:
            func (method): 入口函数
        """
        self._get_sub_byds_func = func

    def get_sub_byds(self, ds_name) -> list:
        """
        返回对ds_name的订阅列表
        """
        if self._get_sub_byds_func:
            return self._get_sub_byds_func(ds_name)
        return []

    def set_get_sub_byds_withsid_func(self, func):
        """
        框架设置获取按ds获取订阅列表函数,包含sid信息
        Parameters:
            func (method): 入口函数
        """
        self._get_sub_byds_withsid_func = func

    def get_sub_byds_withsid(self, ds_name) -> list:
        """
        返回根据sid多对ds_name的订阅列表
        """
        if self._get_sub_byds_withsid_func:
            return self._get_sub_byds_withsid_func(ds_name)
        return []


class TaskUnit(UnitBase):
    """
    这是说明
    """

    def __init__(self):
        super().__init__()
        self._mutex_current_step = threading.Lock()
        self._dic_current_step = {}
        self.update_step_struct(['initializing', 'running', 'done'], '')

    @classmethod
    def i18n_retinfo(cls, ret_info) -> str:
        """
        国际化返回提示信息,负责单元范围内的返回信息国际化,远期将各语言版本分别用对应文件加载
        """
        sup_ret_info = super().i18n_retinfo(ret_info)
        i18n = os.getenv('I18N', 'en')
        i18n_all = {
            'cn': {
            }
        }
        return i18n_all.get(i18n, {}).get(ret_info, sup_ret_info)

    def update_step_struct(self, steps, root_steps) -> None:
        """
        设置进度节点，系统默认提供['initializing','running','done']这三个节点
        示例:
        Parameters:
            steps: 进度子节点
            root_steps: 所属的父节点，层级间用';'隔开
        """
        with self._mutex_current_step:
            point_step = self._dic_current_step
            if len(root_steps) > 0:
                roots = root_steps.split(';')
                for level in roots:
                    point_step = point_step['sub_steps'][level]
            point_step['time_start'] = time.time()
            point_step['time_cost'] = 0
            point_step['current_step'] = 0
            point_step['total_step'] = len(steps)
            point_step['current_info'] = steps[0]
            point_step['time_remain'] = 10
            point_step['steps'] = steps
            point_step['sub_steps'] = {}
            for sub_step in steps:
                point_step['sub_steps'][sub_step] = {}

    def _fresh_step(self, step_layer) -> None:
        """
        刷新step信息
        Parameters:
            step_layer: 对应的step的列表
        """
        sub_time_remain = 0
        if 'sub_steps' in step_layer:
            for sub in step_layer['sub_steps']:
                if len(step_layer['sub_steps'][sub]) > 0:
                    self._fresh_step(step_layer['sub_steps'][sub])
                if 'time_remain' in step_layer['sub_steps'][sub]:
                    sub_time_remain += step_layer['sub_steps'][sub]['time_remain']
        if 'time_remain' in step_layer:
            step_layer['time_remain'] = max(
                step_layer['time_remain'], sub_time_remain)
        if 'time_start' in step_layer and 'time_cost' in step_layer:
            step_layer['time_cost'] = time.time()-step_layer['time_start']

    def update_current_step(self, current_step, root_steps) -> None:
        """
        更新当前进度
        示例:
        当前进度为  'running;获取期货公司列表;step1',调用方法为:self.update_current_step('step1','runnning;获取期货公司列表')
        Parameters:
            current_step: 当前进度节点
            root_steps: 所属根节点，层级间用';'隔开
        """
        with self._mutex_current_step:
            point_step = self._dic_current_step
            if len(root_steps) > 0:
                roots = root_steps.split(';')
                for level in roots:
                    point_step = point_step['sub_steps'][level]
            cur_index = point_step['steps'].index(current_step)
            if cur_index == 0:
                point_step['time_start'] = time.time()
            else:
                point_step['current_step'] = cur_index
                point_step['current_info'] = current_step
                point_step['time_cost'] = time.time()-point_step['time_start']
                avgtime_step = point_step['time_cost']/cur_index
                point_step['time_remain'] = avgtime_step * \
                    (point_step['total_step']-cur_index-1)
                self._fresh_step(self._dic_current_step)
            self.logger.info('%s', self._dic_current_step)

    def req_get(self, url, params=None, **kwargs):
        """
        封装reuest.get,调用频率控制,缓存,错误打印,ip池等
        """
        resp = requests.get(url, params=params, **kwargs)
        time.sleep(random.uniform(0.5, 3))
        if resp.status_code != 200:
            self.logger.error('Get request [%s] failed, code[%d], info[%s]',
                              url, resp.status_code, resp.text)
        return resp

    def req_post(self, url, data=None, json=None, **kwargs):
        """
        封装reuest.post,调用频率控制,缓存,错误打印,ip池等
        """
        resp = requests.post(url, data=data, json=json, **kwargs)
        time.sleep(random.uniform(0.5, 3))
        if resp.status_code != 200:
            self.logger.error('Post request [%s] failed, code[%d], info[%s]',
                              url, resp.status_code, resp.text)
        return resp

    @abstractmethod
    def run(self) -> tuple:
        """
        执行task unit任务

        Returns:
            code: 执行结果代码
            info: 执行结果信息
            content: 执行结果
        """
        return (0, '', '')


class TradeStrategyUnit(UnitBase):
    """
    这是说明
    """

    def __init__(self):
        super().__init__()
        self._user = ''

    def hello(self):
        """
        这是说明
        """
        self._user = 'username'

    def hello2(self):
        """
        这是说明
        """
        self._user = 'username'
